#!/usr/bin/env python

APPEND = "append"
APPEND_CONST = "append_const"
COUNT = "count"
EXTEND = "extend"
HELP = "help"
PARSERS = "parsers"
STORE = "store"
STORE_CONST = "store_const"
STORE_FALSE = "store_false"
STORE_TRUE = "store_true"
VERSION = "version"
